import json
import logging
import types

import requests
import typedecorator

from urconf.uptimerobot_syncable import Contact, Monitor

logger = logging.getLogger("urconf")
typedecorator.setup_typecheck()

DEFAULT_INTERVAL = 5  # minutes


class UptimeRobotAPIError(Exception):
    """An exception which is raised when Uptime Robot API returns an error."""
    pass


class UptimeRobot(object):
    """UptimeRobot is the main object used to sync configuration.

    It keeps alert contacts and monitors defined by the user in
    `self._contacts` and `self_monitors` lists.
    """
    @typedecorator.params(self=object, api_key=str, url=str, dry_run=bool)
    def __init__(self, api_key, url="https://api.uptimerobot.com/v2/",
                 dry_run=False):
        """Initializes the configuration.

        Args:
            api_key: (string) Uptime Robot API key. This should be the
                "Main API Key", not one of the monitor-specific API keys.
            url: (string) Base URL for Uptime Robot API.
            dry_run: (bool) Flag that can be set to True to prevent urconf
                from changing Uptime Robot configuration.
        """
        self._url = url.rstrip("/") + "/"
        self._dry_run = dry_run
        # These are HTTP query parameters that will be passed to the API with
        # all requests.
        self.params = {
            "api_key": api_key,
            "format": "json",
        }
        self._contacts = {}
        self._monitors = {}
        # `requests` logs at INFO by default, which is annoying.
        logging.getLogger("requests").setLevel(logging.WARNING)

    @typedecorator.params(self=object, method=str,
                          params={str: typedecorator.Union(str, int)})
    def _api_post(self, method, params):
        """Issues a POST request to the API and returns the result.

        Args:
            method: (string) API method to call.
            params: ({string: string}) A dictionary containing key/value
                pairs that will be used in POST data.

        Returns:
            Unmarshalled API response as a Python object.

        Raises:
            UptimeRobotAPIError: when API returns an unexpected error.
        """
        url = self._url + method
        resp = requests.post(url, data=params)
        if resp.status_code != 200:
            raise UptimeRobotAPIError("Got HTTP error {} fetching {}".format(
                resp.status_code, url))
        logger.debug("POST {} {}: {}".format(url, params, resp.text))
        try:
            data = json.loads(resp.text)
        except ValueError as e:
            raise UptimeRobotAPIError(
                "Error decoding JSON of {}: {}. Got: {}".format(
                    method, e, resp.text))
        if data["stat"] != "ok":
            raise UptimeRobotAPIError("{} returned error: {}".format(
                method, data["error"]))
        return data

    @typedecorator.params(
        self=object, method=str, params={str: typedecorator.Union(str, int)},
        element_func=types.FunctionType)
    def _api_post_paginated(self, method, params, element_func):
        """Fetches all elements from a given API method.

        This function gets all elements that a given API method returns,
        issuing multiple POST calls if results do not fit in a single page.

        Args:
            method: (string) API method to call.
            params: ({string: string}) A dictionary containing key/value
                pairs that will be used in the URL query string.
            element_func: function that extracts a list of results from the
                object returned by the API call in question. For example, if
                returned JSON is `{"result": [...]}`, the function can be
                `lambda x: x["result"]`.

        Returns:
            A list of Python objects corresponding to API response.
        """
        params = params.copy()
        result = []
        while True:
            response = self._api_post(method, params)
            result.extend(element_func(response))
            if "pagination" in response:
                response = response["pagination"]
            if response["total"] > response["offset"] + response["limit"]:
                params["offset"] = response["offset"] + response["limit"]
            else:
                break
        return result

    def _sync_monitors(self):
        """Synchronizes locally defined list of monitors with the server.

        This method compares locally defined monitors with the result of the
        `getMonitors` API method and synchronizes them by creating missing
        monitors, deleting obsolete ones, and updating the ones that changed.

        Note: creating and updating monitors requires server-side contact IDs,
        so `_sync_monitors` should only be executed after `_sync_contacts`.
        """
        existing = {}
        params = self.params.copy()
        params.update({"alert_contacts": 1})
        fetched = self._api_post_paginated(
            "getMonitors", params, lambda x: x["monitors"])
        for monitor_dict in fetched:
            m = Monitor(**monitor_dict)
            if m.name in self._monitors:
                existing[m.name] = True
                if not m == self._monitors[m.name]:
                    self._api_update_monitor(m, self._monitors[m.name])
            else:
                self._api_delete_monitor(m)
        for name in self._monitors:
            if name not in existing:
                self._api_create_monitor(self._monitors[name])

    @typedecorator.params(self=object, old="Monitor", new="Monitor")
    def _api_update_monitor(self, old, new):
        logger.info("Updating monitor {}".format(new.name))
        logger.debug("Old: %s", old)
        logger.debug("New: %s", new)
        if old["type"] != new["type"]:
            logger.info("Monitor type updates are not possible, "
                        "will remove and re-add {}".format(new.name))
            self._api_delete_monitor(old)
            self._api_create_monitor(new)
            return
        if self._dry_run:
            return
        params = self.params.copy()
        params.update(new._params_update)
        params["id"] = old["id"]
        self._api_post("editMonitor", params)

    @typedecorator.params(self=object, monitor="Monitor")
    def _api_delete_monitor(self, monitor):
        logger.info("Deleting monitor {}".format(monitor.name))
        if self._dry_run:
            return
        params = self.params.copy()
        params.update(monitor._params_delete)
        self._api_post("deleteMonitor", params)

    @typedecorator.params(self=object, monitor="Monitor")
    def _api_create_monitor(self, monitor):
        logger.info("Creating monitor {}".format(monitor.name))
        if self._dry_run:
            return
        params = self.params.copy()
        params.update(monitor._params_create)
        self._api_post("newMonitor", params)

    def _sync_contacts(self):
        """Synchronizes locally defined list of contacts with the server.

        This method compares locally defined contacts with the result of the
        `getAlertContacts` API method and synchronizes them by creating missing
        contacts and deleting obsolete ones.

        This also populates server-side contact IDs that are required to create
        and update monitors, so `_sync_contacts` should be executed before
        `_sync_monitors`.
        """
        existing = {}
        fetched = self._api_post_paginated(
            "getAlertContacts", self.params,
            lambda x: x["alert_contacts"])
        for contact_dict in fetched:
            c = Contact(**contact_dict)
            if c.name in self._contacts:
                existing[c.name] = True
                # Populate the `id` field based on the contact information
                # we got from the server. This id will be required for the
                # newMonitor / editMonitor calls we make later.
                self._contacts[c.name]["id"] = c["id"]
                if c != self._contacts[c.name]:
                    self._api_update_contact(c, self._contacts[c.name])
            else:
                self._api_delete_contact(c)
        for name in self._contacts:
            if name not in existing:
                contact_id = self._api_create_contact(self._contacts[name])
                self._contacts[name]["id"] = contact_id

    @typedecorator.params(self=object, old="Contact", new="Contact")
    def _api_update_contact(self, old, new):
        logger.info("Updating contact {}".format(new.name))
        logger.debug("Old: %s", old)
        logger.debug("New: %s", new)
        if old["type"] != new["type"]:
            logger.info("Contact type updates are not possible, "
                        "will remove and re-add {}".format(new.name))
            self._api_delete_contact(old)
            self._api_create_contact(new)
            return
        if self._dry_run:
            return
        params = self.params.copy()
        params.update(new._params_update)
        params["id"] = old["id"]
        self._api_post("editAlertContact", params)

    @typedecorator.params(self=object, contact="Contact")
    def _api_delete_contact(self, contact):
        logger.info("Deleting contact {}".format(contact.name))
        if self._dry_run:
            return
        params = self.params.copy()
        params.update(contact._params_delete)
        self._api_post("deleteAlertContact", params)

    @typedecorator.params(self=object, contact="Contact")
    def _api_create_contact(self, contact):
        logger.info("Creating contact {}".format(contact.name))
        if self._dry_run:
            return
        params = self.params.copy()
        params.update(contact._params_create)
        result = self._api_post("newAlertContact", params)
        return result["alertcontact"]["id"]

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, type=int, value=str, friendlyname=str)
    def contact(self, type, value, friendlyname=""):
        """Adds a contact of a given type.

        This is mainly a convenience function for other type-specific methods
        (like email_contact), however it can be used directly to define a
        contact of a type which is not supported by the newAlertContact
        endpoint of Uptime Robot API. The contact will need to be created in
        the Uptime Robot UI manually, but using this function to define it
        as part of configuration will allow usage of such contact for
        monitors.

        Args:
            type: (int) contact type.
            value: (string) contact value.
            friendlyname: (string) name used for this contact in the Uptime
                Robot.

        Returns:
            Contact object which can later be used in add_contacts method
                of a monitor.
        """
        c = Contact(friendly_name=friendlyname, type=type, value=value)
        assert c.name not in self._contacts, \
            "Duplicate contact: {}".format(c.name)
        self._contacts[c.name] = c
        logging.debug("Created contact: %s", c)
        return c

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, email=str, name=str)
    def email_contact(self, email, name=None):
        """Defines an email contact.

        Args:
            email: (string) e-mail address.
            name: (string) name used for this contact in the Uptime Robot web
                interface.

        Returns:
            Contact object which can later be used in add_contacts method
                of a monitor.
        """
        if not name:
            name = email
        return self.contact(Contact.TYPE_EMAIL, email, name)

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, key=str, name=str)
    def boxcar_contact(self, key, name=""):
        return self.contact(Contact.TYPE_BOXCAR, key, name)

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, number=str, name=str)
    def sms_contact(self, number, name=""):
        return self.contact(Contact.TYPE_SMS, number, name)

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, value=str, name=str)
    def twitter_dm_contact(self, value, name=""):
        return self.contact(Contact.TYPE_TWITTER_DM, value, name)

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, value=str, name=str)
    def webhook_contact(self, value, name=""):
        return self.contact(Contact.TYPE_WEBHOOK, value, name)

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, value=str, name=str)
    def pushbullet_contact(self, value, name=""):
        return self.contact(Contact.TYPE_PUSHBULLET, value, name)

    @typedecorator.returns("Contact")
    @typedecorator.params(self=object, value=str, name=str)
    def pushover_contact(self, value, name=""):
        return self.contact(Contact.TYPE_PUSHOVER, value, name)

    @typedecorator.returns("Monitor")
    @typedecorator.params(
        self=object, name=str, url=str, keyword=str, should_exist=bool,
        http_username=str, http_password=str, interval=int)
    def keyword_monitor(self, name, url, keyword, should_exist=True,
                        http_username="", http_password="",
                        interval=DEFAULT_INTERVAL):
        """Defines a keyword monitor.

        Args:
            name: (string) name used for this monitor in the Uptime Robot web
                interface.
            url: (string) URL to check.
            keyword: (string) Keyword to check.
            should_exist: (string) Whether the keyword should exist or not
                (defaults to True).
            http_username: (string) Username to use for HTTP authentification.
            http_password: (string) Password to use for HTTP authentification.
            interval: (int) Monitoring interval in minutes.

        Returns:
            Monitor object.
        """
        keywordtype = 2 if should_exist else 1
        m = Monitor(friendly_name=name, type=Monitor.TYPE_KEYWORD, url=url,
                    keyword_value=keyword, keyword_type=keywordtype,
                    http_username=http_username, http_password=http_password,
                    interval=interval*60)
        assert m.name not in self._monitors, \
            "Duplicate monitor: {}".format(m.name)
        self._monitors[m.name] = m
        return m

    @typedecorator.returns("Monitor")
    @typedecorator.params(self=object, name=str, hostname=str, port=int,
                          interval=int)
    def port_monitor(self, name, hostname, port, interval=DEFAULT_INTERVAL):
        """Defines a port monitor.

        Args:
            name: (string) name used for this monitor in the Uptime Robot web
                interface.
            hostname: (string) Host name to check.
            port: (int) TCP port.
            interval: (int) Monitoring interval in minutes.

        Returns:
            Monitor object.
        """
        # Port to subtype map from https://uptimerobot.com/api
        port_to_subtype = {80: 1, 443: 2, 21: 3, 25: 4, 110: 5, 143: 6}
        subtype = port_to_subtype.setdefault(port, 99)
        m = Monitor(friendly_name=name, type=Monitor.TYPE_PORT, url=hostname,
                    sub_type=subtype, port=port, interval=interval*60)
        assert m.name not in self._monitors, \
            "Duplicate monitor: {}".format(m.name)
        self._monitors[m.name] = m
        return m

    def sync(self):
        """Synchronizes configuration with the Uptime Robot API.

        This method should be called after all contacts and monitors have been
        defined and will sync defined configuration to the Uptime Robot."""
        self._sync_contacts()
        self._sync_monitors()



import pytest

from urconf.uptimerobot_syncable import Contact, Monitor


class TestUptimeRobotSyncable(object):
    def test_required_fields(self):
        with pytest.raises(RuntimeError):
            Contact(friendly_name="name")

    def test_id_does_not_affect_equality(self):
        contact1 = Contact(friendly_name="c1", type=2, value="v1", id="0213")
        contact2 = Contact(friendly_name="c1", type=2, value="v1", id="1444")
        assert contact1 == contact2
        # __repr__ includes `id`, so string representations are not equal.
        assert str(contact1) != str(contact2)

    def test_monitor_contacts_affect_equality(self):
        contact = Contact(friendly_name="c1", type=2, value="v1", id="0213")
        mon1 = Monitor(friendly_name="m1", url="u1", type="1")
        mon1.add_contacts(contact)
        mon2 = Monitor(friendly_name="m1", url="u1", type="1",
                       alert_contacts=[
                           {"id": "0213", "type": 2, "value": "v1",
                           "threshold": 0, "recurrence": 0}])
        mon3 = Monitor(friendly_name="m1", url="u1", type=1)  # no contacts
        assert mon1 != mon3
        assert mon2 != mon3
        assert mon1 == mon2
        assert str(mon1) == str(mon1)



import os
try:
    from urllib.parse import urlparse, parse_qs
except ImportError:
    from urlparse import urlparse, parse_qs

import pytest
import requests
import responses
import urconf


def read_data(filename):
    """Reads data from file and returns it as a string.

    File will be read from the directory named after this file (test_foo/ if
    the file is test_foo.py).
    """
    basename, _ = os.path.splitext(__file__)
    with open(os.path.join(basename, filename)) as f:
        return f.read()


def assert_query_params(request, **kwargs):
    """Asserts that given query parameters have expected values.

    Args:
        request: a PreparedRequest object.
        kwargs: key/value pairs that should be present in the query of
            request URL.

    Raises:
        AssertionError if a parameter has unexpected value.
        KeyError if a parameter does not exist.
    """
    params = parse_qs(request.body, keep_blank_values=True)
    for key in kwargs:
        assert params[key][0] == str(kwargs[key]), "Invalid {}".format(key)


class TestUptimeRobot(object):
    @responses.activate
    def test_get_raises_on_invalid_json(self):
        responses.add(
            responses.POST, "https://fake/none", body="omg this is not json")

        config = urconf.UptimeRobot("", url="https://fake")
        with pytest.raises(urconf.uptimerobot.UptimeRobotAPIError):
            config._api_post("none", {})

    @responses.activate
    def test_get_raises_on_404(self):
        responses.add(
            responses.POST, "https://fake/none", body="404", status=404)

        config = urconf.UptimeRobot("", url="https://fake")
        with pytest.raises(urconf.uptimerobot.UptimeRobotAPIError):
            config._api_post("none", {})

    @responses.activate
    def test_get_raises_on_api_errors(self):
        responses.add(responses.POST, "https://fake/none",
                      body='{"stat":"fail","error":{"type":"invalid_parameter"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        with pytest.raises(urconf.uptimerobot.UptimeRobotAPIError):
            config._api_post("none", {})

    @responses.activate
    def test_api_post_paginated(self):
        def callback(request):
            params = parse_qs(request.body) if request.body else {}
            limit = params["limit"][0] if "limit" in params else 1
            offset = params["offset"][0] if "offset" in params else 0
            resp = """{{"stat": "ok", "offset": {offset}, "limit": {limit},
                        "total": 10,"fake":["fakedata{offset}"]}}""".format(
                offset=offset, limit=limit)
            return (200, {}, resp)
        responses.add_callback(responses.POST, "https://fake/getFake",
                               callback=callback)

        config = urconf.UptimeRobot("", url="https://fake/")
        result = config._api_post_paginated("getFake", {}, lambda x: x["fake"])

        assert len(responses.calls) == 10
        for i in (range(10)):
            assert "fakedata{}".format(i) in result

    @responses.activate
    def test_add_email_contact(self):
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_one"))
        responses.add(
            responses.POST, "https://fake/newAlertContact",
            body='{"stat": "ok","alertcontact":{"id":"0725","status":0}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        config.email_contact("e@mail", name="email1")
        config.email_contact("XYZ")
        config._sync_contacts()

        assert config._contacts["XYZ"]["id"] == "0725"
        assert len(responses.calls) == 2
        assert_query_params(
            responses.calls[1].request, type=2,
            friendly_name="XYZ", value="XYZ")

    @responses.activate
    def test_add_boxcar_contact(self):
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_one"))
        responses.add(
            responses.POST, "https://fake/newAlertContact",
            body='{"stat": "ok","alertcontact":{"id":"12344","status":"0"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        config.email_contact("e@mail", name="email1")
        config.boxcar_contact("XYZ", name="boxcar1")
        config._sync_contacts()

        assert config._contacts["XYZ"]["id"] == "12344"
        assert len(responses.calls) == 2
        assert_query_params(
            responses.calls[1].request, type=4,
            friendly_name="boxcar1", value="XYZ")

    @responses.activate
    def test_delete_email_contact(self):
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_two"))
        responses.add(
            responses.POST, "https://fake/deleteAlertContact",
            body='{"stat": "ok","alert_contact":{"id":"9876352"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        config.email_contact("e@mail", name="email1")
        config._sync_contacts()

        assert len(responses.calls) == 2
        assert_query_params(
            responses.calls[1].request, id="9876352")

    @responses.activate
    def test_add_port_monitor(self):
        responses.add(responses.POST, "https://fake/getMonitors",
                      body=read_data("monitors_none"))
        responses.add(
            responses.POST, "https://fake/newMonitor",
            body='{"stat": "ok","monitor":{"id":"515","status":1}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        config.port_monitor("my mail", "servername", 25),
        config._sync_monitors()

        assert len(responses.calls) == 2
        assert_query_params(
            responses.calls[1].request, friendly_name="my mail",
            url="servername", type=4, sub_type=4,
            port=25, alert_contacts="",
            interval=urconf.uptimerobot.DEFAULT_INTERVAL*60)

    @responses.activate
    def test_add_keyword_monitor_and_change_contact_threshold(self):
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_one"))
        responses.add(responses.POST, "https://fake/getMonitors",
                      body=read_data("monitors_three"))
        responses.add(
            responses.POST, "https://fake/editMonitor",
            body='{"stat": "ok","monitor":{"id":"123403"}}')
        responses.add(
            responses.POST, "https://fake/newMonitor",
            body='{"stat": "ok","monitor":{"id":"6969","status":"1"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        email = config.email_contact("e@mail", name="email1")
        config.keyword_monitor(
            "kw1", "http://fake", "test1", http_username="user1",
            http_password="pass1").add_contacts(email)
        config.port_monitor("ssh1", "host1", 22).add_contacts(email)
        config.port_monitor(
            "smtp2", "host2", 25).add_contacts(email, threshold=5)
        config.keyword_monitor(
            "kw2", "http://fake2", "test2").add_contacts(email)
        config.sync()

        assert len(responses.calls) == 4
        assert_query_params(
            responses.calls[2].request, friendly_name="smtp2",
            url="host2", sub_type=4,
            alert_contacts="012345_5_0",
            interval=urconf.uptimerobot.DEFAULT_INTERVAL*60)
        assert_query_params(
            responses.calls[3].request, friendly_name="kw2",
            url="http://fake2",
            keyword_type=2, keyword_value="test2",
            alert_contacts="012345_0_0",
            interval=urconf.uptimerobot.DEFAULT_INTERVAL*60)

    @responses.activate
    def test_remove_monitor(self):
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_one"))
        responses.add(responses.POST, "https://fake/getMonitors",
                      body=read_data("monitors_three"))
        responses.add(
            responses.POST, "https://fake/deleteMonitor",
            body='{"stat": "ok","monitor":{"id":"123403"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        email = config.email_contact("e@mail", name="email1")
        config.port_monitor("ssh1", "host1", 22).add_contacts(email)
        config.port_monitor("smtp2", "host2", 25).add_contacts(email)
        config.sync()

        assert len(responses.calls) == 3
        assert_query_params(responses.calls[2].request, id=123401)

    @responses.activate
    def test_edit_monitor_type(self):
        """API does not allow editing monitor type, so urconf should
           remove the old monitor and create the new one.
        """
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_one"))
        responses.add(responses.POST, "https://fake/getMonitors",
                      body=read_data("monitors_three"))
        responses.add(
            responses.POST, "https://fake/deleteMonitor",
            body='{"stat": "ok","monitor":{"id":"123403"}}')
        responses.add(
            responses.POST, "https://fake/newMonitor",
            body='{"stat": "ok","monitor":{"id":"120011","status":"1"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        email = config.email_contact("e@mail", name="email1")
        # change keyword monitor to a port monitor
        config.port_monitor("kw1", "fake", 80).add_contacts(email)
        config.port_monitor("ssh1", "host1", 22).add_contacts(email)
        config.port_monitor("smtp2", "host2", 25).add_contacts(email)
        config.sync()

        assert len(responses.calls) == 4
        assert_query_params(responses.calls[2].request, id=123401)
        assert_query_params(
            responses.calls[3].request, friendly_name="kw1",
            url="fake", type=4, sub_type=1,
            alert_contacts="012345_0_0",
            interval=urconf.uptimerobot.DEFAULT_INTERVAL*60)

    @responses.activate
    def test_remove_http_auth(self):
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_one"))
        responses.add(responses.POST, "https://fake/getMonitors",
                      body=read_data("monitors_three"))
        responses.add(
            responses.POST, "https://fake/editMonitor",
            body='{"stat": "ok","monitor":{"id":"123401"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        email = config.email_contact("e@mail", name="email1")
        config.keyword_monitor(
            "kw1", "http://fake", "test1").add_contacts(email)
        config.port_monitor("ssh1", "host1", 22).add_contacts(email)
        config.port_monitor("smtp2", "host2", 25).add_contacts(email)
        config.sync()

        assert len(responses.calls) == 3
        assert_query_params(
            responses.calls[2].request, friendly_name="kw1",
            url="http://fake",
            keyword_type=2, keyword_value="test1",
            alert_contacts="012345_0_0",
            interval=urconf.uptimerobot.DEFAULT_INTERVAL*60)

    @responses.activate
    def test_change_email_address(self):
        """Tests contact updates.

        Since API does not allow editing contact type, this verifies that the
        contact gets removed and then re-added. New contact ID will be
        allocated, so all monitors using the old contact will need to be
        updated as well.
        """
        responses.add(responses.POST, "https://fake/getAlertContacts",
                      body=read_data("contacts_one"))
        responses.add(
            responses.POST, "https://fake/deleteAlertContact",
            body='{"stat": "ok","alertcontact":{"id":"012345"}}')
        responses.add(
            responses.POST, "https://fake/newAlertContact",
            body='{"stat": "ok","alertcontact":{"id":"144444","status":"0"}}')
        responses.add(responses.POST, "https://fake/getMonitors",
                      body=read_data("monitors_three"))
        responses.add(
            responses.POST, "https://fake/editMonitor",
            body='{"stat": "ok","monitor":{"id":"123401"}}')
        responses.add(
            responses.POST, "https://fake/editMonitor",
            body='{"stat": "ok","monitor":{"id":"123402"}}')
        responses.add(
            responses.POST, "https://fake/editMonitor",
            body='{"stat": "ok","monitor":{"id":"123403"}}')

        config = urconf.UptimeRobot("", url="https://fake/")
        email = config.boxcar_contact("boxcar1", name="email1")
        config.keyword_monitor(
            "kw1", "http://fake", "test1", http_username="user1",
            http_password="pass1").add_contacts(email)
        config.port_monitor("ssh1", "host1", 22).add_contacts(email)
        config.port_monitor("smtp2", "host2", 25).add_contacts(email)
        config.sync()

        assert len(responses.calls) == 7
        assert_query_params(
            responses.calls[1].request, id="012345")
        assert_query_params(
            responses.calls[2].request,
            friendly_name="email1", type=4, value="boxcar1")
        assert_query_params(
            responses.calls[4].request, friendly_name="kw1",
            url="http://fake",
            keyword_type=2, keyword_value="test1",
            http_username="user1", http_password="pass1",
            alert_contacts="144444_0_0",
            interval=urconf.uptimerobot.DEFAULT_INTERVAL*60)

    @responses.activate
    def test_change_email_address_dry_run(self):
        """Tests dry run mode, confirming that no objects get changed."""
        with responses.RequestsMock(assert_all_requests_are_fired=False) \
                as resp:
            resp.add(responses.POST, "https://fake/getAlertContacts",
                     body=read_data("contacts_two"))
            resp.add(responses.POST, "https://fake/getMonitors",
                     body=read_data("monitors_three"))
            exception = requests.exceptions.HTTPError(
                "dry_run should not mutate state")
            for method in ("deleteAlertContact", "newAlertContact",
                           "editMonitor", "deleteMonitor", "newMonitor"):
                resp.add(responses.POST, "https://fake/{}".format(method),
                         body=exception)

            config = urconf.UptimeRobot("", url="https://fake/", dry_run=True)
            email = config.email_contact("new@mail", name="email2")
            config.keyword_monitor(
                "kw1", "http://fake", "test1").add_contacts(email)
            config.port_monitor("ssh1", "host1", 22).add_contacts(email)
            config.port_monitor("smtp3", "host3", 25).add_contacts(email)
            config.sync()

            assert len(resp.calls) == 2



from setuptools import setup, find_packages

setup(
    name="urconf",
    version="2021.1",
    url="http://github.com/knyar/urconf",
    license="MIT",
    author="Anton Tolchanov",
    description="Declarative configuration library for Uptime Robot",
    long_description=open("README.rst", "r").read(),
    packages=find_packages(),
    platforms="any",
    install_requires=["requests", "typedecorator"],
    keywords=["monitoring", "api", "uptime robot"],
)



class Syncable(object):
    """Syncable is the parent class for objects that urconf keeps in sync.

       All values are stored in the `self._values` dictionary.
    """
    def __init__(self, **kwargs):
        """Construct the object.

        This is designed to be called either with kwargs specified manually or
        with a dict of parameters as returned by getAlertContacts or
        getMonitors.

        Parameter names (self._values keys) correspond to the "Parameters"
        list at https://uptimerobot.com/api
        """
        self._values = {}
        for f in self._REQUIRED_FIELDS:
            if f not in kwargs:
                raise RuntimeError("Contact requires {}; got {}".format(
                    f, kwargs))

        # `id` is not part of FIELDS, because it"s auto-generated on the server
        # rather than passed by the user. However, it's useful to have it,
        # so it's added into the self._values if it exists in kwargs.
        for f in self._FIELDS + ["id"]:
            if f in kwargs and kwargs[f]:
                self[f] = kwargs[f]

    def __eq__(self, other):
        for f in self._FIELDS:
            if self[f] != other[f]:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __setitem__(self, key, value):
        self._values[key] = self._TYPES[key](value)

    def __getitem__(self, key):
        if key in self._values:
            return self._values[key]
        if self._TYPES[key] == str:
            return ""
        if self._TYPES[key] == int:
            return 0

    def __repr__(self):
        return self._values.__repr__()

    @property
    def name(self):
        """Defines primary identificator for this object used by urconf."""
        return self[self._FIELDS[0]]

    @property
    def _params_delete(self):
        """Generates URL parameters for the delete* API call."""
        return {"id": self["id"]}

class Monitor(Syncable):
    _FIELDS = [
        "friendly_name", "url", "type", "sub_type", "keyword_type",
        "keyword_value", "http_username", "http_password", "port", "interval",
        "http_auth_type", "http_method", "post_type", "post_value",
        "post_content_type",
    ]
    _TYPES = {
        # leading zeroes matter, so `id` should be treated is a string.
        "id": str,
        "friendly_name": str,
        "url": str,
        "type": int,
        "sub_type": int,
        "keyword_type": int,
        "keyword_value": str,
        "http_username": str,
        "http_password": str,
        "port": int,
        "interval": int,
        "http_auth_type": int,
        "http_method": int,
        "post_type": int,
        "post_value": str,
        "post_content_type": int,
    }
    _REQUIRED_FIELDS = ["friendly_name", "url", "type"]

    # Possible monitor types, copied from https://uptimerobot.com/api
    TYPE_KEYWORD = 2
    TYPE_PORT = 4

    def __init__(self, **kwargs):
        super(Monitor, self).__init__(**kwargs)
        self._added_contacts = []
        self._contacts_str = None
        if "alert_contacts" in kwargs:
            # This means that this Monitor object has been created based on
            # the data returned by getMonitors API call, which includes contact
            # IDs and options, which can be placed in the right format into
            # self._contacts_str.
            contacts = [
                self._contact_str(c["id"], c["threshold"], c["recurrence"])
                for c in kwargs["alert_contacts"]]
            self._contacts_str = "-".join(sorted(contacts))

    def _contact_str(self, *args):
        return "_".join([str(a) for a in args])

    @property
    def _contacts(self):
        """Returns contact information for this monitor.

        Information is returned in the format expected by editMonitor or
        newMonitor API calls. For monitors that come from the server (i.e.
        initialized based on getMonitors data) we can read the
        self._contacts_str directly. Otherwise we look at all contacts
        added using self.add_contacts.
        """
        if self._contacts_str:
            return self._contacts_str
        contacts = [self._contact_str(c[0]["id"], c[1], c[2])
                    for c in self._added_contacts]
        return "-".join(sorted(contacts))

    def __repr__(self):
        return "<{} {}>".format(self._values, self._contacts)

    def __eq__(self, other):
        if not super(Monitor, self).__eq__(other):
            return False
        return self._contacts == other._contacts

    @property
    def _params_create(self):
        """Generates URL parameters for the newMonitor API call."""
        params = self._values.copy()
        params["alert_contacts"] = self._contacts
        return params

    @property
    def _params_update(self):
        """Generates URL parameters for the editMonitor API call."""
        params = self._params_create
        del params["type"]
        return params

    def add_contacts(self, *args, **kwargs):
        """Defines contacts for a monitor.

        Args:
            args: one or more Contact objects (returned by functions like
                `email_contact` or `boxcar_contact`).
            threshold: alert threshold (the x value that is set to define "if
                down for x minutes, alert every y minutes).
            recurrence: alert recurrence (the y value that is set to define "if
                down for x minutes, alert every y minutes).
        """
        for key in kwargs:
            assert key in ("threshold", "recurrence"), \
                "invalid keyword argument to add_contacts: {}".format(key)
        threshold = kwargs.get("threshold", 0)
        recurrence = kwargs.get("recurrence", 0)
        for c in args:
            assert type(c) == Contact, "{} is not a Contact".format(c)
            self._added_contacts.append((c, threshold, recurrence))


class Contact(Syncable):
    _FIELDS = ["value", "type", "friendly_name"]
    _TYPES = {
        # leading zeroes matter, so `id` should be treated is a string.
        "id": str,
        "friendly_name": str,
        "type": int,
        "value": str,
    }
    _REQUIRED_FIELDS = ["value", "type"]

    # Possible contact types, copied from https://uptimerobot.com/api
    TYPE_SMS = 1
    TYPE_EMAIL = 2
    TYPE_TWITTER_DM = 3
    TYPE_BOXCAR = 4
    TYPE_WEBHOOK = 5
    TYPE_PUSHBULLET = 6
    TYPE_PUSHOVER = 9

    @property
    def _params_create(self):
        """Generates URL parameters for the newAlertContact API call."""
        return {f: self[f] for f in self._FIELDS}

    @property
    def _params_update(self):
        """Generates URL parameters for the editAlertContact API call."""
        params = self._params_create
        del params["type"]
        return params
